/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package ab.demo.other;

import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import ab.demo.InvalidImageException;
import ab.vision.ABObject;
import ab.vision.ABType;
import ab.vision.GameStateExtractor.GameState;
import ab.vision.Vision;
import external.ClientMessageEncoder;

//Java interface of ClientActionRobot
public class ClientActionRobotJava extends ClientActionRobot {

	public ClientActionRobotJava(String ip) {
		super(ip);
	}
   
	//return game state as enum format
	public GameState checkState() {
		byte result = super.getState();
		GameState state = GameState.values()[result];
		return state;
	}
	//return an array of best scores. the nth slot stores the score of (n+1)th level
	public int[] checkScore()
	{
		byte[] scores = super.getBestScores();
		int[] _scores = new int[scores.length/4];
		for (int i = 0 ; i < _scores.length ; i++){
			_scores[i] = super.bytesToInt(scores[i * 4], scores[i*4 + 1], scores[i*4 + 2], scores[i*4 + 3]);
		}
		return _scores;
	}
	
	//send a shot message using int values as input
	public byte[] shoot(int fx, int fy, int dx, int dy, int t1, int t2,
			boolean polar) {
		return super.shoot(intToByteArray(fx), intToByteArray(fy),
				intToByteArray(dx), intToByteArray(dy), intToByteArray(t1),
				intToByteArray(t2), polar);
	}

	//send a shot sequence message using int arrays as input
	//one array per shot
	public byte[] cshootSequence(int[]... shots) {
		byte[][] byteShots = new byte[shots.length][24] ;
		int shotCount = 0;
		for (int[] shot : shots) {
			byteShots[shotCount] = ClientMessageEncoder.mergeArray(intToByteArray(shot[0])/*fx*/,
					intToByteArray(shot[1])/*fy*/, intToByteArray(shot[2])/*dx*/,
					intToByteArray(shot[3])/*dy*/, intToByteArray(shot[4])/*t1*/,
					intToByteArray(shot[5])/*t2*/);
			shotCount++;
		}
		return super.cshootSequence(byteShots);
	}
	public ABType getBirdTypeOnSling() throws InvalidImageException 
	{
		fullyZoomIn();
		BufferedImage screenshot = doScreenShot();
		if (screenshot == null) {
			return ABType.Unknown;
		}
		Vision vision = new Vision(screenshot);
		try {
			Thread.sleep(1000);
		} catch (InterruptedException e) {
			
			e.printStackTrace();
		}
		fullyZoomOut();
		List<ABObject> _birds = vision.findBirdsMBR();
		if(_birds.isEmpty())
			return ABType.Unknown;
		Collections.sort(_birds, new Comparator<Rectangle>(){

			@Override
			public int compare(Rectangle o1, Rectangle o2) {
				
				return ((Integer)(o1.y)).compareTo((Integer)(o2.y));
			}	
		});
		return _birds.get(0).getType();
	}
	public int[] checkMyScore() {
		byte[] scores = super.getMyScore();
		int[] _scores = new int[scores.length/4];
		for (int i = 0 ; i < _scores.length ; i++){
			_scores[i] = super.bytesToInt(scores[i * 4], scores[i*4 + 1], scores[i*4 + 2], scores[i*4 + 3]);
		}
		return _scores;
	}

	public byte[] bid(byte[] amount) {
		byte[] result = new byte[4];
		try {
			out.write(ClientMessageEncoder.bid(amount));
			out.flush();
			
			in.read(result);
			return result;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return result;
	}
	
	public byte[] getCompetitionState() {
		byte[] result = new byte[4];
		try {
			out.write(ClientMessageEncoder.getCompetitionState());
			out.flush();
			
			in.read(result);	
			
			return result;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return result;
	}

	public byte[] getTurn() {
		byte[] result = new byte[4];
		try {
			out.write(ClientMessageEncoder.getTurn());
			out.flush();
			
			in.read(result);
			return result;
		} catch (IOException e) {
			e.printStackTrace();
		}
		return result;
	}

	public int[] getLevels() {
		int level = 3;
		int totalBytes = level * 4;
		byte[] buffer = new byte[totalBytes];
		int[] result = new int[level];
		try {
			out.write(ClientMessageEncoder.getLevels());
			out.flush();
		
			in.read(buffer);
			ByteBuffer bytebuffer = ByteBuffer.wrap(buffer);
			for (int i = 0; i < level; i++) {
				result[i] = bytebuffer.getInt();
			}
		    return result; 
			
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return result;	
	}

	public int getOpponent() {
		byte[] result = new byte[4];
		try {
			out.write(ClientMessageEncoder.getOpponent());
			out.flush();
			
			in.read(result);
			return ByteBuffer.wrap(result).getInt();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return 0;
	}

	public int[] getOpponentScore() {
		int level = 3;
		int totalBytes = level * 4;
		byte[] buffer = new byte[totalBytes];
		int[] result = new int[level];
		try {
			out.write(ClientMessageEncoder.getOpponentScore());
			out.flush();
		
			in.read(buffer);
			ByteBuffer bytebuffer = ByteBuffer.wrap(buffer);
			for (int i = 0; i < level; i++) {
				result[i] = bytebuffer.getInt();
			}
		    return result; 
			
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return result;	
	}
	
	public int[] getMyRoundScore() {
		int level = 3;
		int totalBytes = level * 4;
		byte[] buffer = new byte[totalBytes];
		int[] result = new int[level];
		try {
			out.write(ClientMessageEncoder.getMyRoundScore());
			out.flush();
		
			in.read(buffer);
			ByteBuffer bytebuffer = ByteBuffer.wrap(buffer);
			for (int i = 0; i < level; i++) {
				result[i] = bytebuffer.getInt();
			}
		    return result; 
			
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return result;	
	}

	public  HashMap<Integer, HashMap<Integer, Integer[]>> getScoreBoard() {
		try {
			out.write(ClientMessageEncoder.getScoreBoard());
			out.flush();
			
			byte[] ids = new byte[8*4];
			ArrayList<Integer> result_ids = new ArrayList<Integer>();;
			in.read(ids);
			ByteBuffer bytebuffer = ByteBuffer.wrap(ids);
			int count = 0;
			for (int i = 0; i < 8; i++) {
				int result_id = bytebuffer.getInt();
				if (result_id > 0) {
					result_ids.add(result_id);
					count++;
				}
				System.out.println(result_id);
			}
			int num_rounds = count-1;
			
			// intitialise scoreboard
			byte[] buffer = new byte[7*8*3*4];
			int[] scores_buffer = new int[7*8*3];
			in.read(buffer);
			bytebuffer = ByteBuffer.wrap(buffer);
			for (int i = 0; i < 7*8*3; i++) {
				scores_buffer[i] = bytebuffer.getInt();
			}
			
			HashMap<Integer,HashMap<Integer,Integer[]>> scores = new HashMap<Integer,HashMap<Integer,Integer[]>>();
			for (int i = 0; i < num_rounds; i++) {
				HashMap<Integer,Integer[]> temp = new HashMap<Integer,Integer[]>();
				int n = 0;
				for (Integer id : result_ids) {
						  Integer[] scoreSet1 = {scores_buffer[(i*8*3) + (n*3) + 0],
								  				 scores_buffer[(i*8*3) + (n*3) + 1],
								  				 scores_buffer[(i*8*3) + (n*3) + 2]};
						  temp.put(id-1, scoreSet1);
						  n++;
				}
				scores.put(i, temp);
			}
			//ObjectInputStream ois = new ObjectInputStream(in);
			//HashMap<Integer, HashMap<Integer, Integer[]>> temp = (HashMap<Integer,HashMap<Integer,Integer[]>>)ois.readObject();
			//ois.close();
			return scores;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
		return null;
	}

	public byte wasteBird() {
		try {
			out.write(ClientMessageEncoder.wasteBird());
			out.flush();
			
			byte[] inbuffer = new byte[16];
			
			in.read(inbuffer);
			return inbuffer[0];
		} catch (IOException e) {

			e.printStackTrace();
		}
		return 0;
		
	}
}
