/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package proxy;

import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.SynchronousQueue;

import org.java_websocket.WebSocket;
import org.java_websocket.handshake.ClientHandshake;
import org.java_websocket.server.WebSocketServer;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.json.simple.JSONValue;

import server.util.Log;

public class Proxy extends WebSocketServer {
	// current request id
	private HashMap<WebSocket, Long> ids = new HashMap<WebSocket, Long>();
	// map incoming replies to their original request
	private HashMap<WebSocket, HashMap<Long, Result<?>>> results = new HashMap<WebSocket, HashMap<Long, Result<?>>>();
	
	//each websocket has a unique id, used to find out its counterpart.
	private HashMap<WebSocket, Integer> websocket_id = new HashMap<WebSocket, Integer>();

	
	public LinkedBlockingQueue<WebSocket> connectionQueue = new LinkedBlockingQueue<WebSocket>();
	
	//============= Uncomment if test the shooting delay ==========================
	//private boolean rec = false;
	//private Long last_time = 0L;
	
	
	// a request-reply pair
	private class Result<T> {
		// the command that was associated with the original request
		public ProxyMessage<T> message;
		// a blocking queue to wait for the reply
		public SynchronousQueue<Object> queue = new SynchronousQueue<Object>();
	}
	
	public Proxy(int port) throws UnknownHostException {
		super(new InetSocketAddress(port));
	}

	public Proxy(InetSocketAddress address) {
		super(address);
	}

	@Override
	public void onOpen(WebSocket conn, ClientHandshake handshake) {
		
		ids.put(conn, 0L);
		results.put(conn, new HashMap<Long, Result<?>>());
		connectionQueue.add(conn);
		//assign connection an unique id;
		websocket_id.put(conn, connectionQueue.size());
		onOpen();
	
	}

	@Override
	public void onClose(WebSocket conn, int code, String reason, boolean remote) {
		ids.remove(conn);
		results.remove(conn);
		onClose();
	}

	@Override
	public void onMessage(WebSocket conn, String message) {
		JSONArray j = (JSONArray) JSONValue.parse(message);
		Long id = (Long) j.get(0);
		JSONObject data = (JSONObject) j.get(1);

		// get the request for the reply
		Result<?> result = results.get(conn).get(id);
		
		if (result != null) {
			results.remove(id);
			try {
				// notify the queue that the reply is available
				result.queue.put(result.message.gotResponse(data));
			} catch (InterruptedException e) {
			}
		}
	}

	@Override
	public void onError(WebSocket conn, Exception ex) {
		ex.printStackTrace();
	}


	// send the command to the server
	@SuppressWarnings("unchecked")
	public synchronized <T> T send(WebSocket client, ProxyMessage<T> message) {
		
		//long cost = System.nanoTime(); // Uncomment this for testing shooting delay
		Long id = ids.get(client);
		
		// construct the JSON request
		JSONArray a = new JSONArray();
		a.add(id);
		a.add(message.getMessageName());
		a.add(message.getJSON());

		// prepare a result for when a reply is received
		Result<T> result = new Result<T>();
		result.message = message;
		results.get(client).put(id, result);
		//long _cost = System.nanoTime() - cost; //Uncomment this for testing shooting delay
		
		// broadcast the request to all connected clients\
		
		// Long time = System.nanoTime(); // Uncomment this for testing shooting delay
		client.send(a.toJSONString());

       
		/* Uncomment this for testing shooting delay
		 * if(message instanceof ProxyDragMessage)
		{
			
			System.out.println(" cost: " + _cost);
			if(!rec)
	        { 	
	        	last_time = time;
	        	
	        }
			if(rec)
    			System.out.println(" tap, gap_time " + (time - last_time));
    		else {
				System.out.println("shoot " + time);
			}
			if(!rec)
	    		rec = true;
	    	else 
	    		rec = false;
		}*/
        
		// increment the id for the next message
		ids.put(client, id + 1);
		
		try {
			// block and wait until a reply is received
			return (T)result.queue.take();
		} catch (InterruptedException e) {
			//ABServer.SaveServerState();
			e.printStackTrace();
		}
		return null;
	}
	public void onOpen() {
	}

	public void onClose() {
	}
	
	public synchronized void releaseChromeWindow(WebSocket chrome)
	{
	  Log.sysecho("############### release chrome window  ##############");
	  if(chrome != null)
	  { 
		  boolean returned = false;
		  for (WebSocket available : connectionQueue)
			  if (available.equals(chrome))
			  {
				  System.out.println(chrome.toString() + " was already returned ");
				  returned = true;
			  	  break;
			  }
		  if (!returned)
			  System.out.println(connectionQueue.add(chrome));
		}
	}
	
	// poll the current connections until the specified number of clients has connected
	public synchronized WebSocket waitForClient() {
		WebSocket r = null;
		try {
			r = connectionQueue.take();
		} catch (InterruptedException e) {
		}
		return r;
	}
	public synchronized int checkSocketId(WebSocket websocket)
	{
		return websocket_id.get(websocket);
	}
	//Get a websocket by its id 
	public synchronized WebSocket getWebSocket(int id)
	{
		WebSocket socket = null;
		for (WebSocket _socket: websocket_id.keySet())
		{
			if(websocket_id.get(_socket) == id)
				return _socket;
		}
		return socket;
	}
}