/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package server;

import io.LevelVerificationFileIO;
import io.ScoreIO;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.concurrent.ConcurrentHashMap;

import proxy.Proxy;
import server.schema.LoadingLevelSchema;
import server.util.Log;
import client.logger.Configuration;
import client.logger.PlayerInfo;
import display.DisplayForwardingAgent;

public class ABServer {
	ServerSocket providerSocket;

	Socket connection = null;
	OutputStream out;
	InputStream in;
	private static Proxy agentProxy;
	private static Proxy screenshotProxy;
	public static DisplayForwardingAgent displayAgent;
	public static byte round = 1; // 1: 1st qualification round, 2: 2nd
	public static int group_count = 1;
	public static String file_dir = "";
	public static int current_group = -1;

	// qualification round, 3: final
	// round
	public static byte level = 8;

	public static byte time_limit = 0; // time limit in seconds.
	public static PlayerInfo player_info;
	public static ConcurrentHashMap<Integer, Integer> global = new ConcurrentHashMap<Integer, Integer>();
	public final static ConcurrentHashMap<Integer, Integer> emptyBestscore = new ConcurrentHashMap<Integer, Integer>();
	public static final byte maxLevelNum = 21;

	public static final boolean NO_ABOUT_TO_TERMINATE = false;
	public static String teamScoreFnameFormat = "%dlevelscoreR_%d.txt";
	public static String teamScoreBakFnameFormat = "%dlevelscoreR_%d.bak.%d.txt";
	public static String bestScoreFnameFormat = "bestscoreR_%d.txt";
	public static String bestScoreBakFnameFormat = "bestscoreR_%d.bak.%d.txt";

	public static boolean isLoadPlayerScore() {
		return (round == 2);
	}

	public int start = -1;// in seconds

	ABServer() {
		initialize();
	}

	public static void initialize() {
		player_info = new PlayerInfo();
		global = new ConcurrentHashMap<Integer, Integer>();
		current_group = group_count++;

	}

	ABServer(boolean loadwithstate) {
		if (loadwithstate) {
			LoadServerState();
		} else
			initialize();
		System.out.println(" Time Limit: " + time_limit + " minitues");
		System.out.println(" Level Num: " + level );
		System.out.println(" Round Info: " + round);
		

	}

	void run() {
		
		for (int i = 1; i <= level; i++) {
			global.put(i, 0);
		}

	
		try {

			// 1. creating a server socket
			providerSocket = new ServerSocket(2004, 10);

			// 2. Wait for connection
			System.out.println("AB Server waiting for chrome connection");

			if (agentProxy == null) {
				try {
					agentProxy = new Proxy(9000) {
						@Override
						public void onOpen() {
							System.out
									.println(" A Chrome window is connected to the agent proxy");
						}

						@Override
						public void onClose() {
							System.out
									.println("A Chrome window is disconnected from the agent proxy");
						}
					};
					agentProxy.start();

					System.out
							.println("The Chrome Server (competition agent) started on port: "
									+ agentProxy.getPort());

				} catch (UnknownHostException e) {

					e.printStackTrace();
				}
			}
			if (screenshotProxy == null) {
				try {
					// if port 8888 is not opened, then change to 7777
					screenshotProxy = new Proxy(7777) {
						@Override
						public void onOpen() {
							System.out
									.println(" A Chrome window is connected to the video recording proxy");
						}

						@Override
						public void onClose() {
							System.out
									.println("A Chrome window is disconnected from the video recording proxy");
						}
					};
					screenshotProxy.start();

					System.out
							.println("The Chrome Server (video recording) started on port: "
									+ screenshotProxy.getPort());

				} catch (UnknownHostException e) {

					e.printStackTrace();
				}
			}
			// Initialize scores
			//System.out.println("Round_Info: " + ABServer.round);
			if (round == 1 || round == 2) {
				System.out.println("Load Best Scores");
				int scores[] = ScoreIO.loadScore(ABServer.file_dir);
				// System.out.println(scores.length);
				for (int i = 0; i < scores.length; i++)
					global.put((i + 1), scores[i]);
				for (int i = 0; i < scores.length; i++) {
					System.out.print(" level " + (i + 1) + " : "
							+ global.get((i + 1)) + "  ");
				}
			}
			//System.out.println("Time Limit: " + time_limit + " mins");

			System.out.println("Load Level Verification File: "
					+ "levelVerificationR_" + round + ".txt");
			LoadingLevelSchema.setLevel_Verification(LevelVerificationFileIO
					.readLevelVerificationFile(round));
			ABServerManager ABM = new ABServerManager();
			Thread t_1 = new Thread(ABM);
			t_1.start();

			displayAgent = new DisplayForwardingAgent();
			Thread displayThread = new Thread(displayAgent);
			displayThread.start();

			while (true) {
				Agent agent;
				System.out.println(" waiting for agent  ");
				Socket agentSocket = providerSocket.accept();
				agent = new Agent(agentSocket, agentProxy, screenshotProxy);
				System.out.println(" agent connection established");
				ABM.createAgentThread(agent);
			}
		} catch (IOException ioException) {
			// Close the socket when encounter this exception
			ioException.printStackTrace();
			if (!ABServer.isReset())// if the server was not reset by
									// ABM::dropClient
				ABServer.SaveServerState();
		} finally {
			// 4: Closing connection
			try {
				in.close();
				out.close();
				providerSocket.close();
			} catch (IOException ioException) {
				ioException.printStackTrace();
				if (!ABServer.isReset())// if the server was not reset by
										// ABM::dropClient
					ABServer.SaveServerState();
			}
		}
	}

	private static boolean isReset() {
		return global.isEmpty();
	}

	public synchronized static void updateGlobalScore(int level, int score,
			int player) {

		if (round > 2) {
			
			if (!global.containsKey(level)) {
				global.put(level, score);
			} else if (!(global.get(level) > score)) {
				global.put(level, score);

			}
		}
	}

	protected static void SaveServerState() {
		// ======== Print the Log (Debug Mode)==============
		try {
			File file = new File("serverbackup");
			ObjectOutputStream bos = new ObjectOutputStream(
					new FileOutputStream(file));
			bos.writeObject(player_info);
			bos.writeObject(global);
			//bos.writeObject(AgentLoggerManager.getBmlmap());
			bos.close();
			BufferedWriter buos = new BufferedWriter(new FileWriter("."
					+ File.separator + "DB" + File.separator
					+ "Total_Score_Group_" + current_group));
			buos.write("==============  Group ==========================\n");
			buos.append(" Time elapsed: "
					+ (System.currentTimeMillis() - ABServerManager.current_time)
					/ 1000 + "secs");
			buos.append(player_info.toString_DescendingOrder());
			buos.append("============= Group End ======================\n");
			buos.close();
			Log.sysecho(" Server State has been backed up");
		} catch (FileNotFoundException e) {
			System.out.println("serverBackup not found");
		} catch (IOException e) {
			System.out.println(" Fail to save server state");
		}
		// Save Best Score
		// String filename = "bestscoreR_" + round;
		File file = new File(String.format(bestScoreFnameFormat, round));
		if (file.exists()) {
			File newfile = new File(String.format(bestScoreBakFnameFormat,
					round, System.currentTimeMillis()));
			file.renameTo(newfile);
		}
		try {
			file.createNewFile();
		} catch (IOException e) {

			e.printStackTrace();
		}
		ScoreIO.saveGlobalScore(file);

		// Save Team Score
		for (Configuration conf : player_info.getPlayers_Info().values()) {

			File teamScoreFile = new File(String.format(teamScoreFnameFormat,
					conf.getPlayer_id(), round));

			if (teamScoreFile.exists()) {
				File newfile = new File(String.format(teamScoreBakFnameFormat,
						conf.getPlayer_id(), round, System.currentTimeMillis()));
				teamScoreFile.renameTo(newfile);
			}
			try { 
				teamScoreFile.createNewFile();
			} catch (IOException e) {
				e.printStackTrace();
			}
			ScoreIO.saveTeamScore(teamScoreFile, conf.getScore());
		}

	}

	private static void LoadServerState() {
		/*
		 * @SuppressWarnings("unchecked") try { File file = new
		 * File("serverbackup"); ObjectInputStream bis = new
		 * ObjectInputStream(new FileInputStream( file)); player_info =
		 * (PlayerInfo) bis.readObject();
		 * player_info.getPlayers_Info().get("RMIT_Redbacks")
		 * .setCurrent_level(8); global = (ConcurrentHashMap<Integer, Integer>)
		 * bis.readObject();
		 * 
		 * 
		 * BenchmarkLoggerManager .setBmlmap((HashMap<Integer, BenchmarkLogger>)
		 * bis .readObject());
		 * System.out.println(" The Serever recovery completed"); bis.close() }
		 * catch (FileNotFoundException e) { // TODO Auto-generated catch block
		 * System.out.println("severbackup  not found"); } catch (IOException e)
		 * { System.out.println(" fail to read object"); } catch
		 * (ClassNotFoundException e) { // TODO Auto-generated catch block
		 * System.out.println(" fail to read object, no class found"); }
		 */
	}

	public static String[] extract(String[] commands) {
		if (commands.length < 2)
			return null;
		else {
			String _option = commands[0];
			String _value = commands[1];

			if (_option.equalsIgnoreCase("-t")) {
				int value = Integer.parseInt(_value);
				if (value > 127) {
					ABServer.time_limit = 127;

				} else if (value <= 0) {
					ABServer.time_limit = 0;

				} else {
					ABServer.time_limit = (byte) value;

				}
			} else if (_option.equalsIgnoreCase("-r")) {
				int value = Integer.parseInt(_value);
				if (value < 5)
					ABServer.round = (byte) value;
				else
					System.out
							.println(" Invalid Round Info: choose from [0-5]");
			} else if (_option.equalsIgnoreCase("-h")) {
				ABServer.file_dir = _value;
			} else if (_option.equalsIgnoreCase("-l")) {
				int value = Integer.parseInt(_value);
				ABServer.level = (byte) value;
			}
			String[] _commands = new String[commands.length - 2];
			System.arraycopy(commands, 2, _commands, 0, _commands.length);
			extract(_commands);
		}
		return null;

	}

	public static void main(String args[]) {
		boolean loadState = false;
		File file = new File("." + File.separator + "DB");

		// FileUtils.deleteQuietly(file);
		if (!file.isDirectory())
			file.mkdirs();
		if (args.length > 0) {
			if (args[0].equalsIgnoreCase("recover"))
				loadState = true;
			else if (args[0].equalsIgnoreCase("save"))
				ABServer.SaveServerState();
		}
		ABServer server = new ABServer(loadState);

		 //ABServer.round = 3;
		 //ABServer.level = 8;
	
		 ABServer.time_limit = 64;
		 extract(args);
		 ABServer.file_dir = "bestScoreR_" + (ABServer.round - 1) + ".txt";

		server.run();
	}
}