/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package server;

import gui.ControlPanel;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Observable;
import java.util.Observer;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;

import org.java_websocket.WebSocket;

import server.util.Log;
import server.util.Pair;
import vision.GameStateExtractor.GameState;


public class ABServerManager  implements Runnable,Observer {
	enum Competition_State {Connecting, Bidding, Running, PostShot, Loading, Waiting, Finished};
	
	//private static boolean repeat_turns = false;
	//public static Timer turnTimer;
	
	//public static Competition_State competitionState = Competition_State.Connecting;
	
	//Divide to work group. Group information can be read from text file.
	private static LinkedList<AgentThread> agentThreads = new LinkedList<AgentThread>();
	public static long current_time = -1;
	
	private static ArrayList<Pair> agentPairs = new ArrayList<Pair>();
	private static ArrayList<ArrayList<Pair>> agentPairings = new ArrayList<ArrayList<Pair>>();
	private static ArrayList<Game> games = new ArrayList<Game>();
	
	
	public static LinkedList<AgentThread> getAgentThreads() {
		return agentThreads;
	}
	public  AgentThread createAgentThread(Agent agent)
	{
		AgentThread cm = null;
		System.out.println(" " + (agentThreads.size() + 1) + " agents has been connected" );
	  if(agentThreads.size() < 8)
	  {
			cm = new AgentThread(agent);
			agent.setThreadId(cm.getId());
			agentThreads.add(cm);
			ControlPanel.setEntry(cm.getId(), "Connected");
			agent.addObserver(this);
			cm.start();
			
		}
	  	else
			System.out.println(" up to four agents can be supported. Server cannot accept any more agents");
	  return cm;
	}
	public static void resetServer()
	{
		current_time = -1;
		agentThreads = new LinkedList<AgentThread>();
		ControlPanel.resetAndRepaint();
		//In final rounds, agent can only check the scores of their own group
		ABServer.initialize();
	}
	
	/*
	public static void main(String[] args) {
		// test pairing
		for (int i = 0; i < 8; i++) {
			agentThreads.add(null);
		}
		
		int count = 0;
		for (int test = 0; test < 10000; test++) {
			while (!generate_pairings()) {
				count++;
				System.out.println("retrying");
			};
			
		}
		System.out.println("failed: " + count);
	}*/
	
	public static boolean generate_pairings() {
				agentPairings.clear();
		
				// Generate Pairings
				Random rand = new Random();
				ArrayList<Integer> starting_list = new ArrayList<Integer>();
				for (int i = 1; i <= agentThreads.size(); i++) {
					starting_list.add(i);
				}
				
				int start_at = 0;
				if (agentPairs.size() > 0) {
					agentPairings.add(agentPairs);
					start_at = 1;
				}
				
				for (int i = start_at; i < agentThreads.size()-1; i++) {
					ArrayList<Pair> temp = new ArrayList<Pair>();
					ArrayList<Integer> remaining = (ArrayList<Integer>) starting_list.clone();
					
					for (int j = 0; j < agentThreads.size()/2; j++) {
						
						int min = -1;
						int minIndex = 0;
						for (int l = 0; l < remaining.size(); l++) {
							Integer temp_first = remaining.get(l);
							ArrayList<Integer> not_yet_paired_with = (ArrayList<Integer>) remaining.clone();
							not_yet_paired_with.remove((Object)temp_first);
							for (int k = 0; k < i; k++) {
								for (Pair p : agentPairings.get(k)) {
									if (p.first == temp_first) {
										not_yet_paired_with.remove((Object)p.second);
									}
									if (p.second == temp_first) {
										not_yet_paired_with.remove((Object)p.first);
									}
								}
							}
							if (min == -1 || not_yet_paired_with.size() < min) {
								min = not_yet_paired_with.size() ;
								minIndex = l;
							}
						}
						
						int first = remaining.get(minIndex);
						remaining.remove(minIndex);

						
						System.out.print(first + " ");
						
						ArrayList<Integer> not_yet_paired_with = (ArrayList<Integer>) remaining.clone();
						for (int k = 0; k < i; k++) {
							for (Pair p : agentPairings.get(k)) {
								if (p.first == first) {
									not_yet_paired_with.remove((Object)p.second);
								}
								if (p.second == first) {
									not_yet_paired_with.remove((Object)p.first);
								}
							}
						}
						if (not_yet_paired_with.size() == 0) {
							System.out.println("fail");
							return false;
						}
						
						int second_selection = rand.nextInt(not_yet_paired_with.size());
						int second = not_yet_paired_with.get(second_selection);
						remaining.remove((Object)second);
						temp.add(new Pair(first, second));
						System.out.println("P:"+first+"-"+second+" ");
						
					}
					System.out.println();
					agentPairings.add(temp);
				}
				agentPairs = agentPairings.get(0);
				return true;
	}
	
	public static void startTheCompetition()
	{
		if (agentThreads.size() % 2 != 0) {
			System.out.println("uneven number of agents");
			return;
		}
		
		while (!generate_pairings()) {
			System.out.println("pairing conflict: retrying...");
		};
		
		
		// intitialise scoreboard
		scores = new HashMap<Integer,HashMap<Integer,Integer[]>>();
		for (int i = 0; i < agentPairings.size(); i++) {
			HashMap<Integer,Integer[]> temp = new HashMap<Integer,Integer[]>();
			for (Pair p : agentPairs) {
					  Integer[] scoreSet1 = {0, 0, 0};
					  temp.put(p.first-1, scoreSet1);
					  Integer[] scoreSet2 = {0, 0, 0};
					  temp.put(p.second-1, scoreSet2);
			}
			scores.put(i, temp);
		}
		
		// create games
		for (Pair p : agentPairs) {
			Game game = new Game(agentThreads.get(p.first-1), (p.first-1), agentThreads.get(p.second-1), (p.second-1));
			agentThreads.get(p.first-1).setGame(game);
			agentThreads.get(p.second-1).setGame(game);
			games.add(game);
		}
		
		
		//initialized, if current_time != -1, it means the start button is pressed again before termination
		//Thus we use the original time;
		current_time = System.currentTimeMillis();
		for (AgentThread agentThread : agentThreads)
		{
			if(!agentThread.isRunning())
			{
				  agentThread.runAgent();
				  ControlPanel.setEntry(agentThread.getId(), "Running");
				  ServerTimerTask st = new ServerTimerTask(agentThread);
				  Timer stimer = new Timer();
				  stimer.schedule(st, new Date(current_time + ABServer.time_limit * 60000));		  
			 }
		}
		
		
		startGames();
		System.out.println("Start checking for completion");
		checkForCompletion();
	}
	
	public static int pairing = 0;
	
	public static void switchPairs() {
		pairing = pairing + 1;
		ArrayList<WebSocket> chromeWindows = new ArrayList<WebSocket>();
		for (Pair p : agentPairs) {
			  chromeWindows.add(agentThreads.get(p.first-1).getChromeWindow());
		}
		agentPairs = agentPairings.get(pairing);
		int count = 0;
		for (Pair p : agentPairs) {
			agentThreads.get(p.first-1).setChromeWindow(chromeWindows.get(count));
			agentThreads.get(p.second-1).setChromeWindow(chromeWindows.get(count));
			count++;
		}
	}
	
	private static Timer checkTimer;
	
	public static void checkForCompletion() {
		  checkTimer = new Timer();
		  checkTimer.schedule(new TimerTask() {
			  @Override
			  public void run() {
				  boolean completed = true;
				  for (Game game : games) {
					  System.out.println("checking completed" + game.getCompetitionState().toString() );
					  completed = completed & game.getCompetitionState() == Competition_State.Finished;
				  }
				  if (completed) {
						  System.out.println("Finished Pairing: " + pairing);
						  if (pairing >= agentPairings.size() - 1) {
							  updateScoreBoard();
							  System.out.println("competition Finished");
						  } else {
							  updateScoreBoard();
							  switchPairs();
							  resetGames();
							  startGames();
							  checkForCompletion();
						  }
					  
				  } else {
					  checkForCompletion();
				  }
			  }
			}, 5*1000);
	}
	
	
	public static HashMap<Integer,HashMap<Integer,Integer[]>> scores;
	
	public static void printScores() {
		System.out.println("==SCOREBOARD==");
		for (int i = 0; i < agentPairings.size(); i++) {
			System.out.println("Round " + (i));
			for (Integer key : scores.get(i).keySet()) {
				System.out.print("Agent " + key + ": ");
				for (int j = 0; j < 3; j++) {
					System.out.print("Level " + levels[j] + "(" + scores.get(i).get(key)[j] + ") ");
				}
				System.out.println();
			}
		}
		System.out.println("==============");
	}
	
	public static void updateScoreBoard() {
		for (Game game : games) {
			HashMap<Integer,Integer[]> game_scores = game.getScoreBoard();
			for (Integer agent_id : game_scores.keySet()) {
				scores.get(pairing).put(agent_id, game_scores.get(agent_id));
			}
		}
		printScores();
	}
	
	public static void resetGames() {
		games.clear();
		for (Pair p : agentPairs) {
			Game game = new Game(agentThreads.get(p.first-1), (p.first-1), agentThreads.get(p.second-1), (p.second-1));
			agentThreads.get(p.first-1).setGame(game);
			agentThreads.get(p.second-1).setGame(game);
			games.add(game);
		}
	}
	
	public static void startGames() {
			try {
				Thread.sleep(8000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			
		
		  for (Game game : games) {
			  System.out.println("Starting game 1");
			  (new Thread(game)).start();
		  }
		  System.out.println("Games started");

	}

	
	public static Integer[] levels = {1,2,3};
	
	public static void pair(int a, int b) {
		agentPairs.add(new Pair(a,b));
	}
	public static boolean am_i_first(long threadId) {
		for (Pair p : agentPairs) {
			if (agentThreads.get(p.first-1).getId() == threadId) {
				return true;
			}
			if (agentThreads.get(p.second-1).getId() == threadId) {
				//waitForPartner(p);
				return false;
			}
		}
		return false;
	}
	
	public static boolean waitForPartner(long threadId) {
		for (Pair p : agentPairs) {
			if (agentThreads.get(p.second-1).getId() == threadId) {
				if (agentThreads.get(p.first-1).getChromeWindow() != null) {
					agentThreads.get(p.second-1).setChromeWindow(agentThreads.get(p.first-1).getChromeWindow());
					return true;
				} else {
					return false;
				}
			}
		}
		return false;
	}
	
	/*private static void waitForPartner(Pair p) {
		Timer checkPartnerTimer = new Timer();
		checkPartnerTimer.schedule(new TimerTask() {
			  @Override
			  public void run() {
				if (agentThreads.get(p.first-1).hasChromeWindow()) {
					agentThreads.get(p.second-1).setChromeWindow(agentThreads.get(p.first-1).getChromeWindow());
				} else {
					waitForPartner(p);
				}
			  }
			}, 1*1000);
	}*/
	
	public synchronized static void dropClient(long threadID)
	{
		AgentThread thread = null;
		for (AgentThread at: agentThreads)
		{
			if(at.getId() == threadID)
			{	
				thread = at;
				break;
			}
		}
		if(thread != null)
			agentThreads.remove(thread);
		//check whether all threads have been dropped
		if(agentThreads.isEmpty())
		{
			//reset the server
			Log.sysecho("All the clients have been dropped.", "dropClient", "ABServerManager");
			Log.sysecho("Save Server Data.", "dropClient", "ABServerManager");
			ABServer.SaveServerState();
			Log.sysecho("Reset Server", "dropClient", "ABServerManager");
			resetServer();
		}
	}
	
	@Override
	public void run() 
	{

		  ControlPanel controlPanel = new ControlPanel();
		  controlPanel.getFrame().setVisible(true);
		 
     }
	@Override
	public void update(Observable o, Object arg) {
	  System.out.println("yo");
		
       if(o instanceof Agent)
       {
    	   
    	   Agent agent = (Agent)o;
    	   if(arg instanceof Integer)
    	   {
    		   int id = (Integer)arg;
    		   ControlPanel.updateTeamId( agent.getThreadId(), id);
    	   }
       }
	}
	public static void reconnect() {
	
		for (AgentThread agentThread : agentThreads)
		{
			if(!agentThread.isRunning())
			{
				  agentThread.runAgent();
				  ControlPanel.setEntry(agentThread.getId(), "Running");
				  ServerTimerTask st = new ServerTimerTask(agentThread);
				  Timer stimer = new Timer();
				  stimer.schedule(st, new Date(current_time + ABServer.time_limit * 60000));
			 }
		}
	}
	
	public static int getLevel(int current_level) {
		if (current_level < levels.length) {
			return levels[current_level];
		}
		return -1;
	}

}
