/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package server;

import gui.ControlPanel;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Observable;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.imageio.ImageIO;

import org.java_websocket.WebSocket;

import ab.vision.Vision;
import proxy.Proxy;
import proxy.message.ProxyClickMessage;
import proxy.message.ProxyMouseWheelMessage;
import proxy.message.ProxyScreenshotMessage;
import server.ABServerManager.Competition_State;
import server.schema.LoadingLevelSchema;
import server.schema.RestartLevelSchema;
import server.schema.ShootingSchema;
import server.util.Log;
import server.util.ServerMessageEncoder;
import server.util.Shot;
import server.util.StateUtil;
import vision.GameStateExtractor.GameState;
import client.logger.Configuration;
import client.logger.AgentLogger;
import client.logger.AgentLoggerImageBuffering;
import client.logger.AgentLoggerManager;
import client.messages.ClientMessageDecoder;
import client.messages.ClientMessageTable;

public class Agent extends Observable implements Runnable {
	private Socket client;
	private Proxy agentProxy;
	private Proxy screenshotProxy;
	private long threadId;

	public long getThreadId() {
		return threadId;
	}

	public void setThreadId(long threadId) {
		this.threadId = threadId;
	}

	public Proxy getAgentProxy() {
		return agentProxy;
	}

	public WebSocket getChrome_window() {
		return chrome_window;
	}

	private DataOutputStream out;
	private DataInputStream in;
	public int playerId = -1;
	private int current_score;
	private volatile boolean time = true;
	private volatile boolean wait = true;
	private volatile boolean my_turn = false;
	private volatile boolean wait_for_partner = true;
	private long last_shot;
	private boolean has_shot;
	private boolean has_bid;
	private int current_bid = 0;
	private WebSocket chrome_window;
	private WebSocket chrome_window_screenshot;
	private StateUtil stateUtil;
	private boolean aboutToTerminate = false;
	private Game game;
	private Thread buffering_thread;

	// private AgentRequestHandler handler;

	public StateUtil getStateUtil() {
		return stateUtil;
	}

	private ClientMessageDecoder cmdecoder;
	private ServerMessageEncoder sme;
	private AgentLogger logger;
	private AgentLoggerImageBuffering imageBuffering;
	private Configuration currentConfig;
	private ShootingSchema ss;

	public Configuration getCurrentConfiguration() {
		return currentConfig;
	}

	public AgentLogger getLogger() {
		return logger;
	}

	public Agent(Socket client, Proxy agentProxy, Proxy screenshotProxy) {

		this.client = client;
		this.agentProxy = agentProxy;
		this.screenshotProxy = screenshotProxy;
	

	}

	@Override
	public void run() {

		try {

			out = new DataOutputStream(client.getOutputStream());
			in = new DataInputStream(client.getInputStream());
			cmdecoder = new ClientMessageDecoder();
			sme = new ServerMessageEncoder();
			// accept the configuration from the agent, otherwise, keep
			// looping..
			acceptConfiguration();
			logger.addObserver(ABServer.displayAgent);
			setChanged();
			notifyObservers(playerId);

			while (wait) {
				try {
					Thread.sleep(2000);
				} catch (InterruptedException e) 
				{
					e.printStackTrace();
				}
			}
			// try to move to after pairing decided
			if (ABServerManager.am_i_first(threadId)) {
				chrome_window = agentProxy.waitForClient();
				
				// wait for the web browser to connect
				ss = new ShootingSchema(this);

				// get the chrome window for screenshot;
				chrome_window_screenshot = screenshotProxy.getWebSocket(agentProxy
						.checkSocketId(chrome_window));
				// initialize state util
				stateUtil = new StateUtil(screenshotProxy, chrome_window_screenshot);
				// initialize video recoding utilities
				imageBuffering = new AgentLoggerImageBuffering(logger,
						screenshotProxy, chrome_window_screenshot);
				//buffering_thread = new Thread(imageBuffering);
				//buffering_thread.start();
			} else {
				while (!ABServerManager.waitForPartner(threadId)) {
					System.out.println("Waiting for chrome window");
					try {
						Thread.sleep(2000);
					} catch (InterruptedException e) 
					{
						e.printStackTrace();
					}
				}
			}
			//chrome_window = agentProxy.waitForClient();
			System.out.println("assign chrome window to client: " + 	chrome_window.toString() + " client  " + 
					+ playerId);

			mainBody();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void acceptConfiguration() {
		boolean pass = false;
		while (!pass) {
			try {

				int messageCode = in.read();
				ClientMessageTable message = cmdecoder.getMessage(messageCode);
				if (message == ClientMessageTable.configure
						&& !aboutToTerminate) {

					byte[] _message = new byte[5];
					_message[0] = (byte) messageCode;
					if (in.available() > 0) {

						byte[] buf = new byte[1024];
						in.read(buf);
						// id: four ints
						byte[] digit = new byte[4];
						for (int i = 0; i < 4; i++) {
							digit[i] = buf[i];
							_message[i] = buf[i];
						}
						int player_id = cmdecoder.parseId(digit);

						boolean fb = ABServer.player_info.configure(player_id);

						if (fb) 
						{
							playerId = player_id;

							// register the logger2
							logger = AgentLoggerManager
									.createLogger(playerId);
							logger.setTeam_id(playerId);
							logger.recordMessage(_message);
							currentConfig = ABServer.player_info
									.reterieveConf(playerId);
							
		
							pass = true;
							out.write(sme.econdeConfiguration());
							break;
						} 
						else {
							out.write(ServerMessageEncoder.ERR);
							System.err.println(" ERROR: Invalid Player ID "
									+ player_id);
						}

					} else if (message == null)
						out.write(ServerMessageEncoder.ERR);
					else
						out.write(ServerMessageEncoder.TERMINATE);

				}

			} catch (IOException e) 
			{
				terminate(false);
				e.printStackTrace();
				break;
			}
			try {
				Thread.sleep(3000);
			} catch (InterruptedException e) {
				terminate(false);
				e.printStackTrace();
				break;
			}
		}
	}
	
	public byte[] wasteBird() {
		byte[] result = ServerMessageEncoder.ERR;
		
		client.util.TrajectoryPlanner tp = new client.util.TrajectoryPlanner();
		
		byte[] imageBytes = screenshotProxy.send(
				chrome_window_screenshot,
				new ProxyScreenshotMessage());
		try {
			Thread.sleep(1000);
		} catch (InterruptedException e2) {
			// TODO Auto-generated catch block
			e2.printStackTrace();
		}
   	 	BufferedImage bfimage = null;
   	 	try {
			bfimage = ImageIO.read(new ByteArrayInputStream(imageBytes));
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
		Vision vision = new Vision(bfimage);
		
		Rectangle sling = vision.findSlingshotMBR();
		
		while (sling == null && getGameState() == GameState.PLAYING) {
			for (int i = 0; i < 15; i++) {
				agentProxy.send(chrome_window,
						new ProxyMouseWheelMessage(-1));
			}
			try {
				Thread.sleep(2500);
			} catch (InterruptedException e2) {
				// TODO Auto-generated catch block
				e2.printStackTrace();
			}
			imageBytes = screenshotProxy.send(
					chrome_window_screenshot,
					new ProxyScreenshotMessage());
			bfimage = null;
	   	 	try {
				bfimage = ImageIO.read(new ByteArrayInputStream(imageBytes));
			} catch (IOException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
	   
			vision = new Vision(bfimage);
			sling = vision.findSlingshotMBR();
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				
				e.printStackTrace();
			}
		}
		if (getGameState() != GameState.PLAYING) {
			return result;
		}
		
		Point refPoint = tp.getReferencePoint(sling);
		
		LinkedList<Shot> shots = new LinkedList<Shot>();
		shots.add(new Shot(refPoint.x, refPoint.y, 0, 0, 0, 0));

		// Before executing the shots, we start the thread to
		// capture the screen;

		// imageBuffering.setTheSwtich(true);
		//imageBuffering.setTheSwtich(true);

		ss.shoot(shots);
		Log.echo("Shot Done by " + playerId, "Agent", "Shoot");
		last_shot = System.currentTimeMillis();

		result = ServerMessageEncoder.OK;
		has_shot = true;
		


		/*	System.out
					.println(" wait 2 seconds to stop buffering...");
			try {
				Thread.sleep(2000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			//tryRecordScreenshotsAndScore();
			imageBuffering.setTheSwtich(false);
*/
		
		return result;
	}
	
	
	private static String[] ALL_COMMANDS = { "configure", "doScreenShot","loadLevel","restartLevel","cshoot","pshoot",
											 "cFastshoot", "pFastshoot", "shootSeqFast", "clickInCentre",
											 "getState", "getMyScore","fullyZoomOut", "fullyZoomIn",
											 "getCurrentLevel", "getBestScores" , "shootSeq", "bid", "getCompetitionState", "getTurn", "getOpponent", "getLevels",
											 "getOpponentScore", "getScoreBoard", "", "getMyRoundScore"};
	private static String[] BID_COMMANDS = { "", "doScreenShot","","","","",
											 "", "", "", "clickInCentre",
											 "getState", "","fullyZoomOut", "fullyZoomIn",
											 "", "" , "", "bid", "getCompetitionState", "getTurn", "getOpponent", "getLevels",
											 "getOpponentScore", "getScoreBoard", "", "getMyRoundScore"};
	private static String[] RUN_COMMANDS = { "", "doScreenShot","","","cshoot","pshoot",
											 "cFastshoot", "pFastshoot", "", "clickInCentre",
											 "getState", "","fullyZoomOut", "fullyZoomIn",
											 "", "" , "", "", "getCompetitionState", "getTurn", "getOpponent", "getLevels",
											 "getOpponentScore", "getScoreBoard", "wasteBird", "getMyRoundScore"};
	private static String[] WAIT_COMMANDS = { "", "","","","","",
											 "", "", "", "",
											 "getState", "","", "",
											 "", "" , "", "", "getCompetitionState", "getTurn", "getOpponent", "getLevels",
											 "getOpponentScore", "getScoreBoard", "", "getMyRoundScore"};

	private void mainBody() {
		//String[] respondToCommands = ALL_COMMANDS;
		
		// Calculate some bitmasks
		long respondBitmask = 0;
		long respondAllBitmask = 0;
		long respondBidBitmask = 0;
		long respondRunBitmask = 0;
		long respondWaitBitmask = 0;
		
		List<String> ac = Arrays.asList(ALL_COMMANDS);
		List<String> bc = Arrays.asList(BID_COMMANDS);
		List<String> rc = Arrays.asList(RUN_COMMANDS);
		List<String> wc = Arrays.asList(WAIT_COMMANDS);
		
		for (ClientMessageTable m : ClientMessageTable.values()) {
			int i = ClientMessageTable.getValue(m);
			String name = m.toString();
			respondAllBitmask = respondBitmask | ((ac.contains(name) ? 1 : 0) << i);
			respondBidBitmask = respondBidBitmask | ((bc.contains(name) ? 1 : 0) << i);
			respondRunBitmask = respondRunBitmask | ((rc.contains(name) ? 1 : 0) << i);
			respondWaitBitmask = respondWaitBitmask | ((wc.contains(name) ? 1 : 0) << i);
			
		}
		/*for (int i = 0; i < ALL_COMMANDS.length; i++) {
			respondBitmask = respondBitmask | ((ac.contains(ALL_COMMANDS[i]) ? 1 : 0) << i);
			respondBidBitmask = respondBidBitmask | ((bc.contains(ALL_COMMANDS[i]) ? 1 : 0) << i);
			respondRunBitmask = respondRunBitmask | ((rc.contains(ALL_COMMANDS[i]) ? 1 : 0) << i);
			respondWaitBitmask = respondWaitBitmask | ((wc.contains(ALL_COMMANDS[i]) ? 1 : 0) << i);
		}*/
		

		try {	  // long time = System.nanoTime();

			do {
				if (respondBitmask == respondWaitBitmask) {
					Thread.sleep(1000);
				}
				
				//respondToCommands = WAIT_COMMANDS;
				respondBitmask = respondWaitBitmask;
				if (my_turn) {
					if (game.getCompetitionState() == Competition_State.Running) {
						if (!has_shot) {
							//respondToCommands = RUN_COMMANDS;
							respondBitmask = respondRunBitmask;
						}
					} else if (game.getCompetitionState() == Competition_State.Bidding) {
						//respondToCommands = BID_COMMANDS;
						respondBitmask = respondBidBitmask;
					}
				}

				int messageCode = in.read();

				ClientMessageTable message = cmdecoder.getMessage(messageCode);
				

				if (message != null && !aboutToTerminate) {
					if (playerId == -1) {
						System.out.println(" Client " + "NOT CONFIGURED"
								+ " Message: " + message);
					} else {
						// quiet spam
						if (message != ClientMessageTable.getCompetitionState && message != ClientMessageTable.getTurn) {
							System.out.println(" Client " + playerId
									+ " Message: " + message);
						}
					}
					switch (message) {
					case getScoreBoard: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);
						
						if (respond) {
							
							byte[] ids = new byte [8 * 4];
							
							Set<Integer> key_set = ABServerManager.scores.get(0).keySet();
							
							int count = 0;
							for (Integer id : key_set) {
								System.arraycopy(
										sme.intToByteArray(id+1), 0,
										ids, 4 * (count), 4);
								count++;
							}
							// fill zeros
							/*while (count < 8) {
								System.arraycopy(
										sme.intToByteArray(0), 0,
										ids, 4 * (count), 4);
								count++;
							}*/
							
							byte[] scores = new byte[7*8*3*4];
							
							//HashMap<Integer,Integer[]> empty_score = new HashMap<Integer,Integer[]>();
							int round_size = 8*3*4;
							for (int i = 0; i < 7; i++) {
								if (ABServerManager.scores.containsKey(i)) {
									count = 0;
									for (Integer id : key_set) {
										for (int j = 0; j < 3; j++) {
											int score = ABServerManager.scores.get(i).get(id)[j];
											System.arraycopy(
													sme.intToByteArray(score), 0,
													scores, (round_size*i) + (count*3*4) + (j*4), 4);
										}
										count++;
									}
								} else {
									
								}
							}
							
							out.write(ids);
							out.write(scores);
							/*try {
								ObjectOutputStream oout = new ObjectOutputStream(out);
								oout.writeObject(ABServerManager.scores);
							} catch (IOException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}*/
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case wasteBird: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						
						if (respond) {
							game.shot_made();
							byte[] result = ServerMessageEncoder.ERR;
							result[0] = wasteBird()[0];
							out.write(result);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						
						break;
					}
					case getMyRoundScore: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);

						if (respond) {
							
							int me = game.getMe(this);
							
							byte[] scorebytes = new byte[3 * 4];
							
							int count = 0;
							for (Integer score : game.getScoreBoard().get(me)) {
								System.arraycopy(
										sme.intToByteArray(score), 0,
										scorebytes, 4 * (count), 4);
								count++;
							}



							out.write(scorebytes);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}

						break;
					}
					case getOpponentScore: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);

						if (respond) {
							
							int opponent = game.getOpponent(this);
							
							byte[] scorebytes = new byte[3 * 4];
							
							int count = 0;
							for (Integer score : game.getScoreBoard().get(opponent)) {
								System.arraycopy(
										sme.intToByteArray(score), 0,
										scorebytes, 4 * (count), 4);
								count++;
							}



							out.write(scorebytes);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}

						break;
					}
					case getLevels: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);
						
						if (respond) {
							
							byte[] levelbytes = new byte[3 * 4];
							for (Integer level : ABServerManager.levels) {
								System.arraycopy(
										sme.intToByteArray(level), 0,
										levelbytes, 4 * (level - 1), 4);
							}
							out.write(levelbytes);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case getOpponent: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						if (respond) {
							int opponent = game.getOpponent(this);
							out.write(sme.intToByteArray(opponent));
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case getTurn: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						//logger.recordMessage((byte) messageCode);
						if (respond) {
							out.write(sme.intToByteArray(my_turn ? 1 : 2));
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case getCompetitionState: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						//logger.recordMessage((byte) messageCode);
						if (respond) {
							out.write(sme.intToByteArray(game.getCompetitionState().ordinal()));
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case bid: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						byte[] _message = new byte[5];
						_message[0] = (byte) messageCode;

						if (in.available() > 0) {

							byte[] buf = new byte[1024];
							in.read(buf);
							
							for (int i = 0; i < 4; i++) {

								_message[i] = buf[i];
							}
							
							int test = cmdecoder.parseBid(buf);
							

							logger.recordMessage(_message);
							
							if (respond) {
								System.out.println("bid accepted " + test);
								out.write(sme.intToByteArray(1));
								current_bid = test;
								has_bid = true;
							} else {
								out.write(ServerMessageEncoder.ERR);
							}
						} else
							out.write(ServerMessageEncoder.ERR);

						logger.recordMessage(_message);

						break;
					}
					case doScreenShot: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						// add this request to the log
						logger.recordMessage((byte) messageCode);
						if (respond) {
							byte[] imageBytes = screenshotProxy.send(
									chrome_window_screenshot,
									new ProxyScreenshotMessage());
							out.write(sme.encodeScreenShot(imageBytes));
							Thread.sleep(1000);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case configure: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						byte[] _message = new byte[5];
						_message[0] = (byte) messageCode;
						if (in.available() > 0) {

							byte[] buf = new byte[1024];
							in.read(buf);

							for (int i = 0; i < 4; i++) {

								_message[i] = buf[i];
							}

							if (respond) {
								logger.recordMessage(_message);
								out.write(sme.econdeConfiguration());
							} else {
								out.write(ServerMessageEncoder.ERR);
							}
						} else
							out.write(ServerMessageEncoder.ERR);

						logger.recordMessage(_message);

						break;
					}
					case getState: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);
						if (respond) {
							GameState state = stateUtil.checkCurrentState();
							out.write(sme.encodeState(state));
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case loadLevel: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						if (respond) {
							stopImageBuffering();
							int level = -1;
							byte[] _message = new byte[2];
							_message[0] = (byte) messageCode;
	
							if (in.available() > 0) {
								level = in.read();
	
							} else {
								level = currentConfig.getCurrent_level();
							}
							_message[1] = (byte) level;
	
							System.out.println("Client " + playerId
									+ " Load Level: " + level);
				
							tryRecordScreenshotsAndScore();
	
							LoadingLevelSchema lls = new LoadingLevelSchema(this);
	
							boolean load = lls.loadLevel(level);
	
							out.write((load) ? ServerMessageEncoder.OK
									: ServerMessageEncoder.ERR);
							logger.recordMessage(_message);
							logger.resetLog();
						} else {
							int level = -1;
							byte[] _message = new byte[2];
							_message[0] = (byte) messageCode;
	
							if (in.available() > 0) {
								level = in.read();
	
							} else {
								level = currentConfig.getCurrent_level();
							}
							_message[1] = (byte) level;
	
							System.out.println("Client " + playerId
									+ " Load Level: " + level);
							logger.recordMessage(_message);
							out.write(ServerMessageEncoder.ERR);
						}
						break;

					}
					case restartLevel: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						if (respond) {
							stopImageBuffering();
							boolean output = false;
							{
								tryRecordScreenshotsAndScore();
	
								RestartLevelSchema lls = new RestartLevelSchema(
										this);
								output = lls.restartLevel();
	
							}
							logger.resetLog();
							logger.recordMessage((byte) messageCode);
	
							out.write((output) ? ServerMessageEncoder.OK
									: ServerMessageEncoder.ERR);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}

						break;
					}
					case cshoot: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						byte[] result = ServerMessageEncoder.ERR;
						byte[] _message = new byte[25];
						_message[0] = (byte) messageCode;
						if (in.available() > 0) {
							byte[] shot = new byte[24];
							in.read(shot);
							System.arraycopy(shot, 0, _message, 1, shot.length);
							logger.recordMessage(_message);
							if (respond) {
								result[0] = shoot(shot, false, false)[0];
								out.write(result);
							} else {
								out.write(ServerMessageEncoder.ERR);
							}
						} else
							out.write(result);
						break;
					}
					case cFastshoot: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						byte[] result = ServerMessageEncoder.ERR;
						byte[] _message = new byte[25];
						_message[0] = (byte) messageCode;
						if (in.available() > 0) {
							byte[] shot = new byte[24];
							in.read(shot);
							System.arraycopy(shot, 0, _message, 1, shot.length);
							logger.recordMessage(_message);
							if (respond) {
								result[0] = shoot(shot, false, true)[0];
								out.write(result);
							} else {
								out.write(ServerMessageEncoder.ERR);
							}
						} else
							out.write(result);
						break;
					}
					case pFastshoot: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						byte[] result = ServerMessageEncoder.ERR;
						byte[] _message = new byte[25];
						_message[0] = (byte) messageCode;
						if (in.available() > 0) {
							byte[] shot = new byte[24];
							in.read(shot);
							System.arraycopy(shot, 0, _message, 1, shot.length);
							logger.recordMessage(_message);
							if (respond) {
								result[0] = shoot(shot, false, true)[0];
								out.write(result);
							} else {
								out.write(ServerMessageEncoder.ERR);
							}
						} else
							out.write(result);
						// tryRecordScreenshotsAndScore();
						break;
					}
					case shootSeq: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						

						if (in.available() > 0) {
							int numOfShots = in.read();

							byte[] result = new byte[numOfShots];
							byte[] _message = new byte[numOfShots * 25 + 2];
							_message[0] = (byte) messageCode;
							_message[1] = (byte) numOfShots;

							byte[] shootMessage = new byte[1024];
							byte[] shot = new byte[24];
							in.read(shootMessage);
							
							if (!respond) {
								out.write(ServerMessageEncoder.ERR);
								break;
							}

							System.arraycopy(shootMessage, 0, _message, 2,
									_message.length - 2);
							logger.recordMessage(_message);

							LinkedList<Shot> shots = new LinkedList<Shot>();
							for (int i = 0; i < numOfShots; i++) {
								message = ClientMessageTable
										.getValue(shootMessage[i * 25]);

								if (message == ClientMessageTable.cshoot
										|| message == ClientMessageTable.cFastshoot) {
									System.arraycopy(shootMessage, i * 25 + 1,
											shot, 0, shot.length);
									Shot _shot = parseShoot(shot, false);
									shots.add(_shot);

								} else if (message == ClientMessageTable.pshoot
										|| message == ClientMessageTable.pFastshoot) {
									System.arraycopy(shootMessage, i * 25 + 1,
											shot, 0, shot.length);
									Shot _shot = parseShoot(shot, true);
									shots.add(_shot);

								} else
									break;
							}
							if (stateUtil.checkCurrentState() == GameState.PLAYING) {

								result = ss.shoot(shots);

								try {
									Thread.sleep(15000);
								} catch (InterruptedException e) {

									e.printStackTrace();
								}
								System.out
										.println(" wait 15 seconds to record the score...");
								tryRecordScreenshotsAndScore();

							} else
								System.out
										.println(" Fail to execute the shot: NOT IN PLAYING STATE");

							out.write(result);

						} else {
							logger.recordMessage((byte) messageCode);
							out.write(ServerMessageEncoder.ERR);
						}
						// tryRecordScreenshotsAndScore();

						break;
					}
					case shootSeqFast: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						

						if (in.available() > 0) {
							int numOfShots = in.read();

							byte[] result = new byte[numOfShots];
							byte[] _message = new byte[numOfShots * 25 + 2];
							_message[0] = (byte) messageCode;
							_message[1] = (byte) numOfShots;

							byte[] shootMessage = new byte[1024];
							byte[] shot = new byte[24];
							in.read(shootMessage);
							
							if (!respond) {
								out.write(ServerMessageEncoder.ERR);
								break;
							}
							
							System.arraycopy(shootMessage, 0, _message, 2,
									_message.length);
							logger.recordMessage(_message);

							LinkedList<Shot> shots = new LinkedList<Shot>();
							for (int i = 0; i < numOfShots; i++) {
								message = ClientMessageTable
										.getValue(shootMessage[i * 25]);

								if (message == ClientMessageTable.cshoot
										|| message == ClientMessageTable.cFastshoot) {
									System.arraycopy(shootMessage, i * 25 + 1,
											shot, 0, shot.length);
									Shot _shot = parseShoot(shot, false);
									shots.add(_shot);

								} else if (message == ClientMessageTable.pshoot
										|| message == ClientMessageTable.pFastshoot) {
									System.arraycopy(shootMessage, i * 25 + 1,
											shot, 0, shot.length);
									Shot _shot = parseShoot(shot, true);
									shots.add(_shot);

								} else
									break;
							}
							if (stateUtil.checkCurrentState() == GameState.PLAYING)
								result = ss.shoot(shots);
							else
								System.out
										.println(" Fail to execute the shot: NOT IN PLAYING STATE");
							out.write(result);

						} else {
							logger.recordMessage((byte) messageCode);
							out.write(ServerMessageEncoder.ERR);
						}
						// tryRecordScreenshotsAndScore();
						break;
					}
					case pshoot: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						byte[] result = ServerMessageEncoder.ERR;
						byte[] _message = new byte[25];
						_message[0] = (byte) messageCode;
						if (in.available() > 0) {
							byte[] shot = new byte[24];
							in.read(shot);
							System.arraycopy(shot, 0, _message, 1, shot.length);
							logger.recordMessage(_message);
							if (respond) {
								result[0] = shoot(shot, true, false)[0];
								out.write(result);
							} else {
								out.write(ServerMessageEncoder.ERR);
							}
						} else
							out.write(result);
						// tryRecordScreenshotsAndScore();
						break;
					}

					case fullyZoomOut: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						if (respond) {
							for (int i = 0; i < 15; i++) {
								agentProxy.send(chrome_window,
										new ProxyMouseWheelMessage(-1));
							}
							logger.recordMessage((byte) messageCode);
							Thread.sleep(2500);
							out.write(ServerMessageEncoder.OK);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case fullyZoomIn: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						if (respond) {
							for (int i = 0; i < 15; i++) {
								agentProxy.send(chrome_window,
										new ProxyMouseWheelMessage(1));
							}
							logger.recordMessage((byte) messageCode);
							Thread.sleep(2500);
							out.write(ServerMessageEncoder.OK);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case getCurrentLevel: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						if (respond) {
							Configuration conf = ABServer.player_info
									.reterieveConf(playerId);
							byte level = (byte) conf.getCurrent_level();
							logger.recordMessage((byte) messageCode);
							out.write(level);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case getBestScores: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);
						if (respond) {
							ConcurrentHashMap<Integer, Integer> scores;
							//if (ABServer.round > 2)
								scores = ABServer.global;
							//else
								//scores = ABServer.emptyBestscore;
								
							byte[] scorebytes = new byte[21 * 4];
							for (Integer level : scores.keySet()) {
								System.arraycopy(
										sme.intToByteArray(scores.get(level)), 0,
										scorebytes, 4 * (level - 1), 4);
							}
							out.write(scorebytes);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}
					case getMyScore: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);
						
						if (respond) {
							HashMap<Integer, Integer> scores = currentConfig
									.getScore();
							byte[] scorebytes = new byte[21 * 4];
							for (Integer level : scores.keySet()) {
								System.arraycopy(
										sme.intToByteArray(scores.get(level)), 0,
										scorebytes, 4 * (level - 1), 4);
							}
							out.write(scorebytes);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;

					}
					case clickInCentre: {
						boolean respond = ((1 << messageCode) & respondBitmask) != 0;
						
						logger.recordMessage((byte) messageCode);
						if (respond) {
							agentProxy.send(chrome_window, new ProxyClickMessage(400,
									240));
							Thread.sleep(1000);
							out.write(ServerMessageEncoder.OK);
						} else {
							out.write(ServerMessageEncoder.ERR);
						}
						break;
					}

					}

				} 
				else if (message == null && !aboutToTerminate)
				{
					System.out.println("unrecoginized message, ID: "
							+ messageCode);
					out.write(ServerMessageEncoder.ERR);
				} else
					{
						Log.sysecho("ABOUT_TO_TERMINATE");
						if(playerId == 18249)
						{
							out.write(0);
						}
						else
							out.write(ServerMessageEncoder.TERMINATE);
					}

			} while (time);
		} catch (IOException e) {
			// e.printStackTrace();
			Log.sysecho(" Client  " + playerId + "  is dropped ");
			// ABServer.SaveServerState();
			terminate(false);
			e.printStackTrace();
		} catch (InterruptedException e) {

			Log.sysecho(" Client  " + playerId + "  is dropped ");
			terminate(false);
			e.printStackTrace();
		} finally {
			System.out.println(" Client " + playerId
					+ " has been terminated");
			terminate(false);
			//ABServer.SaveServerState();
		}
	}

	private void tryRecordScreenshotsAndScore() {

		GameState state = stateUtil.checkCurrentState();

		if (state == GameState.WON) {
			tryRecordScore();
			// BenchmarkLoggerManager.getLogger(current_player).end(conf.getCurrent_level(),current_score);

		} 
		  else if (state == GameState.LOST) {
			  int level = currentConfig.getCurrent_level();
			  logger.notifyLevelUnsolved(level);
		  }
		 

	}

	public void tryRecordScore() {

		current_score = stateUtil.checkCurrentScoreSafemode();
		if(current_score != -1){
		int level = currentConfig.getCurrent_level();
		
		ABServer.updateGlobalScore(level, current_score, playerId);
		currentConfig.updateScore(current_score);
	
		logger.notifyLevelSolved(current_score, currentConfig.getTotalScore(), level);
		}
	}

	private Shot parseShoot(byte[] shootMessage, boolean polar) {
		int[] params = cmdecoder.parseShoot(shootMessage);
		// there should 6 params, the order is: fx, fy , dx , dy , t , dt
		int fx = params[0];
		int fy = params[1];
		int dx = params[2];
		int dy = params[3];
		int t = params[4];
		int dt = params[5];

		if (polar) {
			// convert to the Cartesian coordinates.
			int r = dx;
			int theta = dy / 100;
			// double radians = Math.toRadians(theta);
			dx = -(int) (r * Math.cos(Math.toRadians(theta)));
			dy = (int) (r * Math.sin(Math.toRadians(theta)));
		}
		return new Shot(fx, fy, dx, dy, t, dt);

	}

	private byte[] shoot(byte[] shootMessage, boolean polar, boolean fast) {

		byte[] result = ServerMessageEncoder.ERR;
		if (stateUtil.checkCurrentState() == GameState.PLAYING) {
			game.shot_made();
			int[] params = cmdecoder.parseShoot(shootMessage);

			{
				// there should 6 params, the order is: fx, fy , dx , dy , t ,
				// dt
				int fx = params[0];
				int fy = params[1];
				int dx = params[2];
				int dy = params[3];
				int t = params[4];
				int dt = params[5];

				if (polar) {
					// convert to the Cartesian coordinates.
					int r = dx;
					double theta = dy / 100;
					// double radians = Math.toRadians(theta);
					dx = -(int) (r * Math.cos(Math.toRadians(theta)));
					dy = (int) (r * Math.sin(Math.toRadians(theta)));
				}

				// shoot using Cartesian Coordinates
				{

					LinkedList<Shot> shots = new LinkedList<Shot>();
					shots.add(new Shot(fx, fy, dx, dy, t, dt));

					// Before executing the shots, we start the thread to
					// capture the screen;

					// imageBuffering.setTheSwtich(true);
					imageBuffering.setTheSwtich(true);

					ss.shoot(shots);
					Log.echo("Shot Done by " + playerId, "Agent", "Shoot");
					last_shot = System.currentTimeMillis();

					result = ServerMessageEncoder.OK;
					has_shot = true;

					if (!fast) {

						System.out
								.println(" wait 15 seconds to record the score...");
						try {
							Thread.sleep(15000);
						} catch (InterruptedException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
						tryRecordScreenshotsAndScore();
						imageBuffering.setTheSwtich(false);
					} 

				}
			}
		} else {
			System.out
					.println("Fail to execute the shot: NOT IN PLAYING STATE");
		}
		return result;
	}

	public void aboutToTerminate() {
		aboutToTerminate = true;
		
	}

	private void stopImageBuffering() {
		imageBuffering.setTheSwtich(false);
	}

	public void stopWait() {
		wait = false;
	}
	/**
	 * 
	 * @param externalTerminate: true if called outside agent class
	 */
	public synchronized void terminate(boolean externalTerminate) {
     if((time != false && !aboutToTerminate)||(aboutToTerminate && externalTerminate))
     {
    	 	//Go to level selection page when finish
    	 	stateUtil.GoToLevelSelectionPage();
    	 	System.out.println("Release Chrome Winodw " + chrome_window.toString());
			agentProxy.releaseChromeWindow(chrome_window);
			ControlPanel.returnAvailableSlot(threadId);
			ABServerManager.dropClient(threadId);
	
			try {
				client.close();
			} catch (IOException e) {
				System.out.println(" client" + playerId + " is dropped");
			}
			time = false;
		}
	}

	public int getBid() {
		return current_bid;
	}

	public void setTurn(boolean b) {
		if (b) {
			has_shot = false;
		}
		my_turn = b;
	}

	public boolean getTurn() {
		return my_turn;
	}

	public void setChromeWindow(WebSocket chromeWindow) {
		chrome_window = chromeWindow;
		
		// wait for the web browser to connect
		ss = new ShootingSchema(this);

		// get the chrome window for screenshot;
		chrome_window_screenshot = screenshotProxy.getWebSocket(agentProxy
				.checkSocketId(chrome_window));
		// initialize state util
		stateUtil = new StateUtil(screenshotProxy, chrome_window_screenshot);
		// initialize video recoding utilities
		imageBuffering = new AgentLoggerImageBuffering(logger,
				screenshotProxy, chrome_window_screenshot);
		
		//if (buffering_thread != null && buffering_thread.isAlive()) {
//			buffering_thread.interrupt();
	//	}
//		buffering_thread = new Thread(imageBuffering);
	//	buffering_thread.start();
	}

	public void stopWaitForPartner() {
		wait_for_partner = false;
	}
	
	public boolean loadLevel(Integer level) {
		LoadingLevelSchema lls = new LoadingLevelSchema(this);
		
		boolean load = lls.loadLevel(level);
		return load;
	}
	
	public void GoToLevelSelectionPage () {
		stateUtil.GoToLevelSelectionPage();
	}

	public int checkCurrentScoreSafemode() {
		return stateUtil.checkCurrentScoreSafemode();
	}
	
	public GameState getGameState() {
		return stateUtil.checkCurrentState();
	}
	
	public long getLastShotTime() {
		return last_shot;
	}

	public void setGame(Game game) {
		this.game = game;
	}

	public boolean has_bid() {
		return has_bid;
	}
	
	public void new_bid() {
		has_bid = false;
		current_bid = 0;
	}



}
