/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package server.schema;

import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;

import org.java_websocket.WebSocket;

import proxy.Proxy;
import proxy.message.ProxyDragMessage;

import server.Agent;
import server.util.Shot;

public class ShootingSchema {

	private boolean shootImmediately = true;
	private WebSocket chrome_client;
	private Proxy proxy;
	public ShootingSchema(Agent agent) {
		
		chrome_client = agent.getChrome_window();
		proxy = agent.getAgentProxy();
	}
	private boolean isValid(Shot shot)
	{
		int start_x = shot.getX();
		int start_y = shot.getY();
		int end_x = shot.getX() + shot.getDx();
		int end_y = shot.getY() + shot.getDy();
		if ((start_x > 600 || end_x > 600) || (start_y < 100 || end_y < 100))
			return false;
		return true;
	}
	public byte[] shoot(List<Shot> csc) 
	{
		 LinkedList<Shot> shots = new LinkedList<Shot>();
		 byte[] result = new byte[csc.size()];
		int count = 0;
		for (Shot shot : csc) {
			
			//Check the validity of shots
			if(!isValid(shot))
				result[count] = 0;
			else{
					if (shot.getT_shot() != 0)
						shootImmediately = false;
					if (shootImmediately) {
						int t_shot = 5000 * count;
						shots.add(new Shot(shot.getX(), shot.getY(), shot.getDx(), shot
								.getDy(), t_shot));
						if (shot.getT_tap() > 0)
							shots.add(new Shot(0, 0, 0, 0, shot.getT_tap() + t_shot));
					} else {
						shots.add(new Shot(shot.getX(), shot.getY(), shot.getDx(), shot
								.getDy(), shot.getT_shot()));
						if (shot.getT_tap() > 0)
							shots.add(new Shot(0, 0, 0, 0, shot.getT_tap()
									+ shot.getT_shot()));
					}
				result[count] = 1;
			}
			count++;
		}

		Collections.sort(shots, new Comparator<Shot>() {

			@Override
			public int compare(Shot arg0, Shot arg1) {

				return ((Integer) arg0.getT_shot()).compareTo(arg1.getT_shot());
			}
		});
		if (!shots.isEmpty()) {

			  int start_time = shots.getFirst().getT_shot();
		       //Optimize for one shot one time..
			   if(shots.size() < 3)
			   {
				   Shot shot = shots.getFirst();
				   if(shots.size() == 2)
				   {
					   Shot _shot = shots.getLast();
					   int wait_time = (_shot.getT_shot() - start_time)==0?start_time:(_shot.getT_shot()- start_time);
					  //System.out.println(" original wait_time: " + wait_time);
					   long gap = System.currentTimeMillis();
					   proxy.send(chrome_client , new ProxyDragMessage(shot.getX(),shot.getY(),shot.getDx(),shot.getDy()));
					  long await_time = System.nanoTime();
					   gap = System.currentTimeMillis() - gap;
					   wait_time -= gap;
					   if(wait_time < 0)
						   wait_time = 0;
			           
					   try {
						   Thread.sleep(wait_time);
					   } catch (InterruptedException e) {
						
					e.printStackTrace();
					}

					  await_time = System.nanoTime() - await_time;
					  proxy.send(chrome_client, new ProxyDragMessage(_shot.getX(),_shot.getY(),_shot.getDx(),_shot.getDy()));

				   }
				   else {
					   proxy.send(chrome_client, new ProxyDragMessage(shot.getX(),shot.getY(),shot.getDx(),shot.getDy()));
				}
				   result[0] = 1;
				   
			   }
			   else 
			   {
				   long gap = 0;
				 
				   for(Shot _shot: shots)
				   {
					   long wait_time = (_shot.getT_shot() - start_time - gap) <= 0? 0:(_shot.getT_shot()- start_time - gap);
					   
					   try {
						   Thread.sleep(wait_time);
						} catch (InterruptedException e) {
							e.printStackTrace();
						}
					    start_time =  _shot.getT_shot();
						gap = System.currentTimeMillis();
						proxy.send(chrome_client, new ProxyDragMessage(_shot.getX(),_shot.getY(),_shot.getDx(),_shot.getDy()));
				        gap = System.currentTimeMillis() - gap;
				   }
			   }
		}
		return result;
	}

}
