/*****************************************************************************
 ** ANGRYBIRDS AI AGENT FRAMEWORK
 ** Copyright (c) 2015, XiaoYu (Gary) Ge, Jochen Renz,Stephen Gould,
 **  Sahan Abeyasinghe,Jim Keys, Rohan Verma, Andrew Wang, Peng Zhang
 ** All rights reserved.
**This work is licensed under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
**To view a copy of this license, visit http://www.gnu.org/licenses/
 *****************************************************************************/

package server.util;

import java.awt.image.BufferedImage;
import java.io.ByteArrayInputStream;
import java.io.IOException;

import javax.imageio.ImageIO;

import org.java_websocket.WebSocket;

import proxy.Proxy;
import proxy.message.ProxyClickMessage;
import proxy.message.ProxyScreenshotMessage;
import vision.GameStateExtractor;
import vision.GameStateExtractor.GameState;


public class StateUtil {
	private Proxy proxy;
	private WebSocket chrome_client;
	

	public StateUtil(Proxy proxy, WebSocket chrome_client) {
		super();
		this.proxy = proxy;
		this.chrome_client = chrome_client;
	}
	public GameState checkCurrentState()
	{
		return checkCurrentState(proxy, chrome_client);
	}
	private GameState checkCurrentState(Proxy server,WebSocket chrome_client)
	{
	
		byte[] imageBytes = server.send(chrome_client,new ProxyScreenshotMessage());
        GameState state;

		BufferedImage image = null;
	    try {
	           image = ImageIO.read(new ByteArrayInputStream(imageBytes));
	        } 
	    catch (IOException e) {
	        
	        }
	    
        GameStateExtractor gameStateExtractor = new GameStateExtractor();
        state = gameStateExtractor.getGameState(image);
       
		 return state;
	}
	public int checkCurrentScoreSafemode()
	{
		return checkCurrentScoreSafemode(proxy, chrome_client);
	}
	private int checkCurrentScoreSafemode(Proxy server,WebSocket chrome_client)
	{
		int current_score = -1;
		while (current_score != checkCurrentScore(server,
				chrome_client)) {
			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
			
				e.printStackTrace();
			}
		   if(checkCurrentState(server, chrome_client) == GameState.WON)
		   {	
			   current_score = checkCurrentScore(server,
					chrome_client);
				//System.out.println("Debug Check the score: " + current_score);
			}
		   else
		   {
			   System.out.println(" SHOULD BE IN WON STATE");
			   return -1;
		   }
		}
		return current_score;
	}
	public  int checkCurrentScore()
	{
		return checkCurrentScore(proxy,chrome_client);
	}
	private  int checkCurrentScore(Proxy server,WebSocket chrome_client)
	{
		byte[] imageBytes = server.send(chrome_client,new ProxyScreenshotMessage());
        int score = -1;

		BufferedImage image = null;
	    try {
	           image = ImageIO.read(new ByteArrayInputStream(imageBytes));
	        } 
	    catch (IOException e) {
	            // do something
	        }
	    
        GameStateExtractor gameStateExtractor = new GameStateExtractor();
        GameState state = gameStateExtractor.getGameState(image);
        if (state == GameState.PLAYING)
        	score = gameStateExtractor.getScoreInGame(image);
        else
        	if(state == GameState.WON)
        		score = gameStateExtractor.getScoreEndGame(image);
       if(score == -1)
    	   System.out.println("Game Score unavailable ");
        System.out.println("Game state: " + state + " Score: " + score);
		return score;
	}

	public  void GoFromMainMenuToLevelSelection()
	{
		  //--- go from the main menu to the episode menu
		  GameState state = checkCurrentState(proxy,chrome_client) ;
		while (state == GameState.MAIN_MENU){
		       
			  System.out.println("Go to the Episode Menu");
			  proxy.send(chrome_client, new ProxyClickMessage(305,277));
		       try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		   state = checkCurrentState(proxy,chrome_client) ;
		}
		  state = checkCurrentState(proxy,chrome_client) ;
		  //--- go from the episode menu to the level selection  menu
		  while( state == GameState.EPISODE_MENU)
				{
			       System.out.println("Select the Poached Eggs Episode");
			       proxy.send(chrome_client, new ProxyClickMessage(150,300));
			        state = checkCurrentState(proxy , chrome_client) ;
			        try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						
						e.printStackTrace();
					}
			        state = checkCurrentState(proxy , chrome_client);
			  }
		 
	}
	public  void GoToLevelSelectionPage()
	{
		  //--- go from the main menu to the episode menu
		  GameState state = checkCurrentState(proxy,chrome_client) ;
		while (state == GameState.PLAYING){
		       
			  System.out.println("Go to Level Selection Page");
			  proxy.send(chrome_client, new ProxyClickMessage(28,25));
		       try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		       proxy.send(chrome_client, new ProxyClickMessage(164,28));
		       try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
		   state = checkCurrentState(proxy,chrome_client) ;
		}
		  //--- go from the episode menu to the level selection  menu
		  state = checkCurrentState(proxy,chrome_client) ;
		  while( state == GameState.WON || state == GameState.LOST)
		  {
			       System.out.println("Go to the Poached Eggs Episode");
			       proxy.send(chrome_client, new ProxyClickMessage(342, 382));
			        state = checkCurrentState(proxy , chrome_client) ;
			        try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						
						e.printStackTrace();
					}
			        state = checkCurrentState(proxy , chrome_client);
			  }
		  state = checkCurrentState(proxy , chrome_client);
		  while( state == GameState.MAIN_MENU)
		  {
			  System.out.println("Go to the Episode Menu");
			  proxy.send(chrome_client, new ProxyClickMessage(305,277));
		       try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		   state = checkCurrentState(proxy,chrome_client) ;
		  }
		  state = checkCurrentState(proxy,chrome_client) ;
		  while( state == GameState.EPISODE_MENU)
			{
		       System.out.println("Select the Poached Eggs Episode");
		       proxy.send(chrome_client, new ProxyClickMessage(150,300));
		        state = checkCurrentState(proxy , chrome_client) ;
		        try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {
					
					e.printStackTrace();
				}
		        state = checkCurrentState(proxy , chrome_client);
		  }
		 
	}
}
